@@ -252,6 +252,7 @@ gdm_auth_secure_display (GdmDisplay *d)
     if (SERVER_IS_LOCAL (d)) {
 	    char hostname[1024];
 
+	    hostname[1023] = '\0';
 	    if (gethostname (hostname, 1023) == 0) {
 		    g_free (d->hostname);
 		    d->hostname = g_strdup (hostname);
@@ -286,6 +287,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    if ( ! SERVER_IS_LOCAL (d)) {
 		    char hostname[1024];
 
+		    hostname[1023] = '\0';
 		    if (gethostname (hostname, 1023) == 0 &&
 			strcmp (hostname, d->hostname) != 0) {
 			    if ( ! add_auth_entry (d, af, af_gdm, FamilyLocal,
@@ -382,7 +384,7 @@ try_open_append (const char *file)
 gboolean
 gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
 {
-    const char *authdir;
+    char *authdir;
     gint authfd;
     FILE *af;
     GSList *auths = NULL;
@@ -395,16 +397,22 @@ gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
 
     gdm_debug ("gdm_auth_user_add: Adding cookie for %d", user);
 
+    /* Determine whether UserAuthDir is specified. Otherwise ~user is used */
+    if ( ! ve_string_empty (GdmUserAuthDir) &&
+	strcmp (GdmUserAuthDir, "~") != 0) {
+	    if (strncmp (GdmUserAuthDir, "~/", 2) == 0) {
+		    authdir = g_strconcat (homedir, &GdmUserAuthDir[1], NULL);
+	    } else {
+		    authdir = g_strdup (GdmUserAuthDir);
+	    }
+    } else {
+	    authdir = g_strdup (homedir);
+    }
+
 try_user_add_again:
 
     locked = FALSE;
 
-    /* Determine whether UserAuthDir is specified. Otherwise ~user is used */
-    if ( ! ve_string_empty (GdmUserAuthDir))
-	authdir = GdmUserAuthDir;
-    else
-	authdir = homedir;
-
     umask (077);
 
     if (authdir == NULL)
@@ -432,6 +440,7 @@ try_user_add_again:
 
 	    umask (022);
 
+	    g_free (authdir);
 	    return FALSE;
 	}
 
@@ -472,6 +481,8 @@ try_user_add_again:
 		goto try_user_add_again;
 	}
 
+	g_free (authdir);
+
 	return FALSE; 
     }
 
@@ -514,6 +525,7 @@ try_user_add_again:
 
     umask (022);
 
+    g_free (authdir);
     return ret;
 }
 
