@@ -72,9 +72,10 @@ gint sessions = 0;		/* Number of remote sessions */
 gint flexi_servers = 0;		/* Number of flexi servers */
 uid_t GdmUserId;		/* Userid under which gdm should run */
 gid_t GdmGroupId;		/* Groupid under which gdm should run */
-pid_t extra_process = -1;	/* An extra process.  Used for quickie 
+pid_t extra_process = 0;	/* An extra process.  Used for quickie 
 				   processes, so that they also get whacked */
 int extra_status = 0;		/* Last status from the last extra process */
+pid_t gdm_main_pid = 0;
 gboolean preserve_ld_vars = FALSE; /* Preserve the ld environment variables */
 gboolean no_daemon = FALSE;	/* Do not daemonize */
 
@@ -101,6 +102,7 @@ gchar *GdmLocaleFile = NULL;
 gchar *GdmGnomeDefaultSession = NULL;
 gchar *GdmAutomaticLogin = NULL;
 gboolean GdmAutomaticLoginEnable = FALSE;
+gchar *GdmLocalNoPasswordUsers = NULL;
 gboolean GdmAlwaysRestartServer = FALSE;
 gchar *GdmConfigurator = NULL;
 gboolean GdmConfigAvailable = FALSE;
@@ -119,8 +121,11 @@ gchar *GdmPostSession = NULL;
 gchar *GdmFailsafeXServer = NULL;
 gchar *GdmXKeepsCrashing = NULL;
 gchar *GdmHalt = NULL;
+gchar *GdmHaltReal = NULL;
 gchar *GdmReboot = NULL;
+gchar *GdmRebootReal = NULL;
 gchar *GdmSuspend = NULL;
+gchar *GdmSuspendReal = NULL;
 gchar *GdmServAuthDir = NULL;
 gchar *GdmUserAuthDir = NULL;
 gchar *GdmUserAuthFile = NULL;
@@ -128,7 +133,6 @@ gchar *GdmUserAuthFB = NULL;
 gchar *GdmPidFile = NULL;
 gchar *GdmDefaultPath = NULL;
 gchar *GdmRootPath = NULL;
-gchar *GdmDefaultLocale = NULL;
 gboolean  GdmKillInitClients = FALSE;
 gint  GdmUserMaxFile = 0;
 gint  GdmSessionMaxFile = 0;
@@ -155,6 +159,9 @@ gint GdmTimedLoginDelay = 0;
 gchar *GdmStandardXServer = NULL;
 gint  GdmFlexibleXServers = 5;
 gchar *GdmXnest = NULL;
+int GdmFirstVT = 7;
+gboolean GdmVTAllocation = TRUE;
+
 
 /* set in the main function */
 char **stored_argv = NULL;
@@ -227,6 +234,7 @@ gdm_config_parse (void)
     GdmDisplayInit = gnome_config_get_string (GDM_KEY_INITDIR);
     GdmAutomaticLoginEnable = gnome_config_get_bool (GDM_KEY_AUTOMATICLOGIN_ENABLE);
     GdmAutomaticLogin = gnome_config_get_string (GDM_KEY_AUTOMATICLOGIN);
+    GdmLocalNoPasswordUsers = gnome_config_get_string (GDM_KEY_LOCALNOPASSWORDUSERS);
     GdmAlwaysRestartServer = gnome_config_get_bool (GDM_KEY_ALWAYSRESTARTSERVER);
     GdmGreeter = gnome_config_get_string (GDM_KEY_GREETER);
     GdmRemoteGreeter = gnome_config_get_string (GDM_KEY_REMOTEGREETER);
@@ -249,7 +257,6 @@ gdm_config_parse (void)
     GdmReboot = gnome_config_get_string (GDM_KEY_REBOOT);
     GdmRetryDelay = gnome_config_get_int (GDM_KEY_RETRYDELAY);
     GdmRootPath = gnome_config_get_string (GDM_KEY_ROOTPATH);
-    GdmDefaultLocale = gnome_config_get_string (GDM_KEY_LOCALE);
     GdmServAuthDir = gnome_config_get_string (GDM_KEY_SERVAUTH);
     GdmSessDir = gnome_config_get_string (GDM_KEY_SESSDIR);
     GdmSuspend = gnome_config_get_string (GDM_KEY_SUSPEND);
@@ -304,6 +311,9 @@ gdm_config_parse (void)
     if (ve_string_empty (GdmXnest))
 	    GdmXnest = NULL;
 
+    GdmFirstVT = gnome_config_get_int (GDM_KEY_FIRSTVT);    
+    GdmVTAllocation = gnome_config_get_bool (GDM_KEY_VTALLOCATION);    
+
     GdmDebug = gnome_config_get_bool (GDM_KEY_DEBUG);
 
     gnome_config_pop_prefix();
@@ -579,6 +589,11 @@ gdm_config_parse (void)
 	    gdm_fail (_("%s: The gdm group should not be root. Aborting!"), "gdm_config_parse");
     }
 
+    /* get the actual commands to use */
+    GdmHaltReal = ve_get_first_working_command (GdmHalt, FALSE);
+    GdmRebootReal = ve_get_first_working_command (GdmReboot, FALSE);
+    GdmSuspendReal = ve_get_first_working_command (GdmSuspend, FALSE);
+
     setegid (GdmGroupId);	/* gid remains `gdm' */
     seteuid (GdmUserId);
 
@@ -733,9 +748,9 @@ gdm_daemonify (void)
     close (1);
     close (2);
 
-    open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-    open("/dev/null", O_RDWR); /* open stdout - fd 1 */
-    open("/dev/null", O_RDWR); /* open stderr - fd 2 */
+    gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+    gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+    gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 }
 
 static void 
@@ -774,27 +789,49 @@ gdm_start_first_unborn_local (int delay)
 	}
 }
 
-static void
-final_cleanup (void)
+void
+gdm_final_cleanup (void)
 {
-	GSList *list;
-	sigset_t mask;
+	GSList *list, *li;
 
-	gdm_debug ("final_cleanup");
+	gdm_debug ("gdm_final_cleanup");
 
-	sigemptyset (&mask);
-	sigaddset (&mask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &mask, NULL); 
+	gdm_sigchld_block_push ();
 
 	if (extra_process > 1) {
 		/* we sigterm extra processes, and we
 		 * don't wait */
 		kill (extra_process, SIGTERM);
-		extra_process = -1;
+		extra_process = 0;
 	}
 
+	gdm_sigchld_block_pop ();
+
 	list = g_slist_copy (displays);
-	g_slist_foreach (list, (GFunc) gdm_display_unmanage, NULL);
+	for (li = list; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		/* XDMCP and FLEXI_XNEST are safe to kill
+		 * immediately */
+		if (d->type == TYPE_XDMCP ||
+		    d->type == TYPE_FLEXI_XNEST)
+			gdm_display_unmanage (d);
+	}
+	g_slist_free (list);
+
+	list = g_slist_copy (displays);
+	/* somewhat of a hack to kill last server
+	 * started first.  This mostly makes things end up on
+	 * the right vt */
+	list = g_slist_reverse (list);
+	for (li = list; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		/* HACK! Wait 2 seconds between killing of local servers
+		 * because X is stupid and full of races and will hang my
+		 * keyboard if I don't */
+		if (li != list)
+			sleep (2);
+		gdm_display_unmanage (d);
+	}
 	g_slist_free (list);
 
 	/* Close stuff */
@@ -864,9 +901,9 @@ deal_with_x_crashes (GdmDisplay *d)
 
 		    /* No error checking here - if it's messed the best response
 		    * is to ignore & try to continue */
-		    open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		    open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		    open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		    gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		    gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		    gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 		    argv[0] = GdmXKeepsCrashing;
 		    argv[1] = NULL;
@@ -939,26 +976,6 @@ deal_with_x_crashes (GdmDisplay *d)
     return FALSE;
 }
 
-static gboolean
-bin_executable (const char *command)
-{
-	char **argv;
-
-	if (ve_string_empty (command))
-		return FALSE;
-
-	argv = ve_split (command);
-	if (argv != NULL &&
-	    argv[0] != NULL &&
-	    access (argv[0], X_OK) == 0) {
-		g_strfreev (argv);
-		return TRUE;
-	} else {
-		g_strfreev (argv);
-		return FALSE;
-	}
-}
-
 static gboolean 
 gdm_cleanup_children (void)
 {
@@ -987,7 +1004,7 @@ gdm_cleanup_children (void)
 
     if (pid == extra_process) {
 	    /* an extra process died, yay! */
-	    extra_process = -1;
+	    extra_process = 0;
 	    extra_status = exitstatus;
 	    return TRUE;
     }
@@ -1002,26 +1019,22 @@ gdm_cleanup_children (void)
 	    gdm_debug ("gdm_cleanup_children: Slave crashed, killing it's "
 		       "children");
 
-	    if (d->sesspid > 0) {
-		    if (kill (d->sesspid, SIGTERM) == 0)
+	    if (d->sesspid > 1 &&
+		kill (d->sesspid, SIGTERM) == 0)
 			    ve_waitpid_no_signal (d->sesspid, NULL, 0);
-		    d->sesspid = 0;
-	    }
-	    if (d->greetpid > 0) {
-		    if (kill (d->greetpid, SIGTERM) == 0)
+	    d->sesspid = 0;
+	    if (d->greetpid > 1 &&
+		kill (d->greetpid, SIGTERM) == 0)
 			    ve_waitpid_no_signal (d->greetpid, NULL, 0);
-		    d->greetpid = 0;
-	    }
-	    if (d->chooserpid > 0) {
-		    if (kill (d->chooserpid, SIGTERM) == 0)
+	    d->greetpid = 0;
+	    if (d->chooserpid > 1 &&
+		kill (d->chooserpid, SIGTERM) == 0)
 			    ve_waitpid_no_signal (d->chooserpid, NULL, 0);
-		    d->chooserpid = 0;
-	    }
-	    if (d->servpid > 0) {
-		    if (kill (d->servpid, SIGTERM) == 0)
+	    d->chooserpid = 0;
+	    if (d->servpid > 1 &&
+		kill (d->servpid, SIGTERM) == 0)
 			    ve_waitpid_no_signal (d->servpid, NULL, 0);
-		    d->servpid = 0;
-	    }
+	    d->servpid = 0;
     }
 
     /* null all these, they are not valid most definately */
@@ -1063,15 +1076,15 @@ gdm_cleanup_children (void)
     /* checkout if we can actually do stuff */
     switch (status) {
     case DISPLAY_REBOOT:
-	    if ( ! bin_executable (GdmReboot))
+	    if (GdmRebootReal == NULL)
 		    status = DISPLAY_REMANAGE;
 	    break;
     case DISPLAY_HALT:
-	    if ( ! bin_executable (GdmHalt))
+	    if (GdmHaltReal == NULL)
 		    status = DISPLAY_REMANAGE;
 	    break;
     case DISPLAY_SUSPEND:
-	    if ( ! bin_executable (GdmSuspend))
+	    if (GdmSuspendReal == NULL)
 		    status = DISPLAY_REMANAGE;
 	    break;
     default:
@@ -1098,10 +1111,10 @@ start_autopsy:
     case DISPLAY_REBOOT:	/* Reboot machine */
 	gdm_info (_("gdm_child_action: Master rebooting..."));
 
-	final_cleanup ();
+	gdm_final_cleanup ();
 	chdir ("/");
 
-	argv = ve_split (GdmReboot);
+	argv = ve_split (GdmRebootReal);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Reboot failed: %s"), strerror (errno));
@@ -1113,10 +1126,10 @@ start_autopsy:
     case DISPLAY_HALT:		/* Halt machine */
 	gdm_info (_("gdm_child_action: Master halting..."));
 
-	final_cleanup ();
+	gdm_final_cleanup ();
 	chdir ("/");
 
-	argv = ve_split (GdmHalt);
+	argv = ve_split (GdmHaltReal);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Halt failed: %s"), strerror (errno));
@@ -1128,10 +1141,10 @@ start_autopsy:
     case DISPLAY_SUSPEND:	/* Suspend machine */
 	gdm_info (_("gdm_child_action: Master suspending..."));
 
-	final_cleanup ();
+	gdm_final_cleanup ();
 	chdir ("/");
 
-	argv = ve_split (GdmSuspend);
+	argv = ve_split (GdmSuspendReal);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Suspend failed: %s"), strerror (errno));
@@ -1233,7 +1246,7 @@ static void
 gdm_restart_now (void)
 {
 	gdm_info (_("GDM restarting ..."));
-	final_cleanup ();
+	gdm_final_cleanup ();
 	if (stored_path != NULL)
 		gnome_setenv ("PATH", stored_path, TRUE);
 	execvp (stored_argv[0], stored_argv);
@@ -1270,7 +1283,7 @@ mainloop_sig_callback (int sig, gpointer data)
     case SIGINT:
     case SIGTERM:
       gdm_debug ("mainloop_sig_callback: Got TERM/INT. Going down!");
-      final_cleanup ();
+      gdm_final_cleanup ();
       exit (EXIT_SUCCESS);
       break;
 
@@ -1376,12 +1389,7 @@ ensure_desc_012 (void)
 	 * we of course count on the fact that open
 	 * opens the lowest available descriptor */
 	for (;;) {
-		fd = open("/dev/null", O_RDWR);
-		/* what to do on fail?  I dunno,
-		 * just keep on going is the safest
-		 * bet I suppose */
-		if (fd < 0)
-			break;
+		fd = gdm_open_dev_null (O_RDWR);
 		/* Once we are up to 3, we're beyond stdin,
 		 * stdout and stderr */
 		if (fd >= 3) {
@@ -1400,6 +1408,8 @@ main (int argc, char *argv[])
     poptContext ctx;
     int nextopt;
 
+    gdm_main_pid = getpid ();
+
     /* We here ensure descriptors 0, 1 and 2 */
     ensure_desc_012 ();
 
@@ -1451,7 +1461,7 @@ main (int argc, char *argv[])
 
         if (pf != NULL &&
 	    fscanf (pf, "%d", &pidv) == 1 &&
-	    kill (pidv, 0) != -1 &&
+	    kill (pidv, 0) == 0 &&
 	    linux_only_is_running (pidv)) {
 		/* make sure the pid file doesn't get wiped */
 		GdmPidFile = NULL;
@@ -1471,6 +1481,9 @@ main (int argc, char *argv[])
 	    fprintf (pf, "%d\n", (int)getpid());
 	    fclose (pf);
 	}
+
+	chdir (GdmServAuthDir);
+	umask (022);
     }
     else
 	gdm_daemonify();
@@ -1609,7 +1622,8 @@ write_x_servers (GdmDisplay *d)
 		char **argv;
 		char *command;
 		argv = gdm_server_resolve_command_line
-			(d, FALSE /* resolve_handled */);
+			(d, FALSE /* resolve_handled */,
+			 NULL /* vtarg */);
 		command = g_strjoinv (" ", argv);
 		g_strfreev (argv);
 		fprintf (fp, "%s local %s\n", d->name, command);
@@ -1628,9 +1642,11 @@ send_slave_ack (GdmDisplay *d)
 		not[1] = '\n';
 		write (d->master_notify_fd, not, 2);
 	}
-	if (d->slavepid > 0) {
+	gdm_sigchld_block_push ();
+	if (d->slavepid > 1) {
 		kill (d->slavepid, SIGUSR2);
 	}
+	gdm_sigchld_block_pop ();
 }
 
 static void
@@ -1643,9 +1659,11 @@ send_slave_command (GdmDisplay *d, const char *command)
 		write (d->master_notify_fd, cmd, strlen (cmd));
 		g_free (cmd);
 	}
-	if (d->slavepid > 0) {
+	gdm_sigchld_block_push ();
+	if (d->slavepid > 1) {
 		kill (d->slavepid, SIGUSR2);
 	}
+	gdm_sigchld_block_pop ();
 }
 
 
@@ -1806,9 +1824,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 		d = gdm_display_lookup (slave_pid);
 
 		if (d != NULL) {
-#ifdef __linux__
 			d->vt = vt_num;
-#endif
 			gdm_debug ("Got VT_NUM == %d", vt_num);
 			/* send ack */
 			send_slave_ack (d);
@@ -1980,8 +1996,10 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 		GSList *li;
 		for (li = displays; li != NULL; li = li->next) {
 			GdmDisplay *d = li->data;
-			if (d->greetpid > 0)
+			gdm_sigchld_block_push ();
+			if (d->greetpid > 1)
 				kill (d->greetpid, SIGHUP);
+			gdm_sigchld_block_pop ();
 		}
 	} else if (strncmp (msg, GDM_SOP_WRITE_X_SERVERS " ",
 		            strlen (GDM_SOP_WRITE_X_SERVERS " ")) == 0) {
@@ -2359,8 +2377,10 @@ notify_displays_int (const char *key, int val)
 			gdm_fdprintf (disp->master_notify_fd,
 				      "%c%s %d\n",
 				      GDM_SLAVE_NOTIFY_KEY, key, val);
-			if (disp->slavepid > 0)
+			gdm_sigchld_block_push ();
+			if (disp->slavepid > 1)
 				kill (disp->slavepid, SIGUSR2);
+			gdm_sigchld_block_pop ();
 		}
 	}
 }
@@ -2375,8 +2395,10 @@ notify_displays_string (const char *key, const char *val)
 			gdm_fdprintf (disp->master_notify_fd,
 				      "%c%s %s\n",
 				      GDM_SLAVE_NOTIFY_KEY, key, val);
-			if (disp->slavepid > 0)
+			gdm_sigchld_block_push ();
+			if (disp->slavepid > 1)
 				kill (disp->slavepid, SIGUSR2);
+			gdm_sigchld_block_pop ();
 		}
 	}
 }
@@ -2705,12 +2727,7 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 					 ve_sure_string (disp->xnest_disp));
 			} else {
 				gdm_connection_printf (conn, "%d",
-#ifdef __linux__
-						       disp->vt
-#else
-						       -1
-#endif
-						      );
+						       disp->vt);
 			}
 		}
 		gdm_connection_write (conn, "\n");
