@@ -78,7 +78,9 @@ static gboolean do_restart_greeter = FALSE; /* if this is true, whack the
 					       greeter and try again */
 static gboolean restart_greeter_now = FALSE; /* restart_greeter_when the
 						SIGCHLD hits */
-static gboolean check_notifies_immediately = 0; /* check notifies as they come */
+static int check_notifies_immediately = 0; /* check notifies as they come */
+static gboolean gdm_wait_for_ack = TRUE; /* wait for ack on all messages to
+				      * the daemon */
 static gboolean greeter_disabled = FALSE;
 static gboolean greeter_no_focus = FALSE;
 
@@ -144,11 +146,11 @@ static void     gdm_slave_greeter (void);
 static void     gdm_slave_chooser (void);
 static void     gdm_slave_session_start (void);
 static void     gdm_slave_session_stop (pid_t sesspid);
-static void     gdm_slave_session_cleanup (void);
 static void     gdm_slave_alrm_handler (int sig);
 static void     gdm_slave_term_handler (int sig);
 static void     gdm_slave_child_handler (int sig);
 static void     gdm_slave_usr2_handler (int sig);
+static void     gdm_slave_quick_exit (gint status);
 static void     gdm_slave_exit (gint status, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
 static void     gdm_child_exit (gint status, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
 static gint     gdm_slave_exec_script (GdmDisplay *d, const gchar *dir,
@@ -515,7 +517,7 @@ gdm_slave_run (GdmDisplay *display)
 			       "In the meantime this display will be\n"
 			       "disabled.  Please restart gdm when\n"
 			       "the problem is corrected."));
-		    _exit (DISPLAY_ABORT);
+		    gdm_slave_quick_exit (DISPLAY_ABORT);
 	    }
 	    gdm_slave_send_num (GDM_SOP_XPID, d->servpid);
 
@@ -596,11 +598,10 @@ gdm_slave_run (GdmDisplay *display)
     /* something may have gone wrong, try xfailed, if local (non-flexi),
      * the toplevel loop of death will handle us */ 
     if (d->handled && d->dsp == NULL) {
-	    gdm_server_stop (d);
 	    if (d->type == TYPE_LOCAL)
-		    _exit (DISPLAY_XFAILED);
+		    gdm_slave_quick_exit (DISPLAY_XFAILED);
 	    else
-		    _exit (DISPLAY_ABORT);
+		    gdm_slave_quick_exit (DISPLAY_ABORT);
     }
 
     /* Some sort of a bug foo to make some servers work or whatnot,
@@ -634,7 +635,7 @@ gdm_slave_run (GdmDisplay *display)
 	    /* yay, we now wait for the server to die,
 	     * which will in fact just exit, so
 	     * this code is a little bit too anal */
-	    while (d->servpid != 0) {
+	    while (d->servpid > 0) {
 		    select (0, NULL, NULL, NULL, NULL);
 	    }
 	    return;
@@ -663,9 +664,7 @@ gdm_slave_run (GdmDisplay *display)
 	    gdm_debug ("gdm_slave_run: Automatic login done");
 	    
 	    if (remanage_asap) {
-		    gdm_server_stop (d);
-		    gdm_verify_cleanup (d);
-		    _exit (DISPLAY_REMANAGE);
+		    gdm_slave_quick_exit (DISPLAY_REMANAGE);
 	    }
 
 	    /* return to gdm_slave_start so that the server
@@ -708,9 +707,7 @@ gdm_slave_run (GdmDisplay *display)
 	    gdm_slave_send_string (GDM_SOP_LOGIN, "");
 
 	    if (remanage_asap) {
-		    gdm_server_stop (d);
-		    gdm_verify_cleanup (d);
-		    _exit (DISPLAY_REMANAGE);
+		    gdm_slave_quick_exit (DISPLAY_REMANAGE);
 	    }
 
 	    if (greet) {
@@ -833,15 +830,18 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 
 	if (pid < 0) {
 		/* return left pointer */
-		Cursor xcursor = XCreateFontCursor (d->dsp, GDK_LEFT_PTR);
+		Cursor xcursor;
+
+		/* can't fork, damnit */
+		display->sesspid = 0;
+	       
+		xcursor = XCreateFontCursor (d->dsp, GDK_LEFT_PTR);
 		XDefineCursor (d->dsp,
 			       DefaultRootWindow (d->dsp),
 			       xcursor);
 		XFreeCursor (d->dsp, xcursor);
 		XSync (d->dsp, False);
 
-		/* can't fork, damnit */
-		display->sesspid = 0;
 		return;
 	}
 
@@ -849,6 +849,8 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 		char **argv;
 		/* child */
 
+		setsid ();
+
 		setuid (0);
 		setgid (0);
 
@@ -899,8 +901,8 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 				 "default location."));
 
 		argv = ve_split
-			(EXPANDED_GDMCONFIGDIR
-			 "/gdmconfig --disable-sound --disable-crash-dialog");
+			(EXPANDED_BINDIR
+			 "/gdmsetup --disable-sound --disable-crash-dialog");
 		if (access (argv[0], X_OK) == 0)
 			execv (argv[0], argv);
 
@@ -1111,9 +1113,7 @@ gdm_slave_wait_for_login (void)
 			d->logged_in = FALSE;
 
 			if (remanage_asap) {
-				gdm_server_stop (d);
-				gdm_verify_cleanup (d);
-				_exit (DISPLAY_REMANAGE);
+				gdm_slave_quick_exit (DISPLAY_REMANAGE);
 			}
 
 			greeter_no_focus = FALSE;
@@ -1455,6 +1455,8 @@ gdm_slave_greeter (void)
     switch (pid) {
 	
     case 0:
+	setsid ();
+
 	sigfillset (&mask);
 	sigdelset (&mask, SIGINT);
 	sigdelset (&mask, SIGTERM);
@@ -1612,6 +1614,7 @@ gdm_slave_greeter (void)
 	gdm_child_exit (DISPLAY_ABORT, _("%s: Error starting greeter on display %s"), "gdm_slave_greeter", d->name);
 	
     case -1:
+	d->greetpid = 0;
 	gdm_slave_exit (DISPLAY_ABORT, _("%s: Can't fork gdmgreeter process"), "gdm_slave_greeter");
 	
     default:
@@ -1658,6 +1661,9 @@ gdm_slave_send (const char *str, gboolean wait_for_ack)
 	int i;
 	uid_t old;
 
+	if ( ! gdm_wait_for_ack)
+		wait_for_ack = FALSE;
+
 	if (gdm_in_signal == 0)
 		gdm_debug ("Sending %s", str);
 
@@ -1692,6 +1698,11 @@ gdm_slave_send (const char *str, gboolean wait_for_ack)
 	     i++) {
 		sleep (1);
 	}
+
+	if (wait_for_ack  &&
+	    ! gdm_got_ack &&
+	    gdm_in_signal == 0)
+		gdm_error ("Timeout occured for sending message %s", str);
 }
 
 void
@@ -1813,6 +1824,8 @@ gdm_slave_chooser (void)
 	switch (pid) {
 
 	case 0:
+		setsid ();
+
 		sigfillset (&mask);
 		sigdelset (&mask, SIGINT);
 		sigdelset (&mask, SIGTERM);
@@ -1923,7 +1936,7 @@ gdm_slave_chooser (void)
 				buf[bytes] ='\0';
 			send_chosen_host (d, buf);
 
-			_exit (DISPLAY_CHOSEN);
+			gdm_slave_quick_exit (DISPLAY_CHOSEN);
 		}
 
 		close (p[0]);
@@ -2136,50 +2149,6 @@ find_prog (const char *name, const char *args, char **retpath)
 	return NULL;
 }
 
-/* this is for the unforunate case when something went seriously wrong, the
- * sysadmin's a wanker or the user has an old language setting */
-static char *
-unaliaslang (const char *origlang)
-{
-	FILE *langlist;
-	char curline[256];
-
-	if (ve_string_empty (GdmLocaleFile))
-		return g_strdup (origlang);
-
-	langlist = fopen (GdmLocaleFile, "r");
-
-	if (langlist == NULL)
-		return g_strdup (origlang);
-
-	while (fgets (curline, sizeof (curline), langlist)) {
-		char *name;
-		char *lang;
-
-		if (curline[0] <= ' ' ||
-		    curline[0] == '#')
-			continue;
-
-		name = strtok (curline, " \t\r\n");
-		if (name == NULL)
-			continue;
-
-		lang = strtok (NULL, " \t\r\n");
-		if (lang == NULL)
-			continue;
-
-		if (g_ascii_strcasecmp (name, origlang) == 0) {
-			fclose (langlist);
-			return g_strdup (lang);
-		}
-	}
-
-	fclose (langlist);
-
-	return g_strdup (origlang);
-
-}
-
 static char *
 dequote (const char *in)
 {
@@ -2208,7 +2177,6 @@ session_child_run (struct passwd *pwent,
 		   const char *session,
 		   const char *save_session,
 		   const char *language,
-		   gboolean def_language,
 		   const char *gnome_session,
 		   gboolean usrcfgok,
 		   gboolean savesess,
@@ -2326,11 +2294,18 @@ session_child_run (struct passwd *pwent,
 	else
 		gnome_setenv ("PATH", GdmDefaultPath, TRUE);
 
-	/* Eeeeek, this no lookie as a correct language code, let's
-	 * try unaliasing it */
-	if (strlen (language) < 3 ||
-	    language[2] != '_') {
-		language = unaliaslang (language);
+	/* Eeeeek, this no lookie as a correct language code,
+	 * just use the system default */
+	if ( ! ve_string_empty (language) &&
+	     ! ve_locale_exists (language)) {
+		/* FIXME: give this error message! */
+		/* Must wait till string freeze is over */
+		/* XXX STRING XXX
+		char *msg = g_strdup_printf (_("Language %s does not exist, using %s"),
+					     language, _("System default"));
+		gdm_error_box (d, GTK_MESSAGE_ERROR, msg);
+		*/
+		language = NULL;
 	}
 
 	setpgid (0, 0);
@@ -2357,26 +2332,19 @@ session_child_run (struct passwd *pwent,
 				_("%s: setusercontext() failed for %s. "
 				  "Aborting."), "gdm_slave_session_start",
 				login);
-
-	/* A different language was selected, or taken from the saved
-	 * prefs of the user */
-	if ( ! def_language) {
-		gnome_setenv ("LANG", language, TRUE);
-		gnome_setenv ("GDM_LANG", language, TRUE);
-	}
 #else
 	if (setuid (pwent->pw_uid) < 0) 
 		gdm_child_exit (DISPLAY_REMANAGE,
 				_("gdm_slave_session_start: Could not become %s. Aborting."), login);
-
-	/* Set locale */
-	gnome_setenv ("LANG", language, TRUE);
-	/* Only force GDM_LANG to something if there is other then
-	 * default selected.  Else let the session do whatever it
-	 * does since we're using sys default */
-	if ( ! def_language)
-		gnome_setenv ("GDM_LANG", language, TRUE);
 #endif
+
+	/* Only force GDM_LANG to something if there is other then
+	 * system default selected.  Else let the session do whatever it
+	 * does since we're using sys default */
+	if ( ! ve_string_empty (language)) {
+		gnome_setenv ("LANG", language, TRUE);
+		gnome_setenv ("GDM_LANG", language, TRUE);
+	}
 	
 	chdir (home_dir);
 
@@ -2396,7 +2364,7 @@ session_child_run (struct passwd *pwent,
 		gchar *cfgstr = g_strconcat ("=", home_dir,
 					     "/.gnome2/gdm=/session/lang", NULL);
 		/* we chose the system default language so wipe the lang key */
-		if (def_language)
+		if (ve_string_empty (language))
 			gnome_config_clean_key (cfgstr);
 		else
 			gnome_config_set_string (cfgstr, language);
@@ -2596,7 +2564,6 @@ gdm_slave_session_start (void)
     char *gnome_session = NULL;
     gboolean savesess = FALSE, savelang = FALSE, savegnomesess = FALSE;
     gboolean usrcfgok = FALSE, sessoptok = FALSE, authok = FALSE;
-    gboolean def_language = FALSE;
     const char *home_dir = NULL;
     gboolean home_dir_ok = FALSE;
     time_t session_start_time, end_time; 
@@ -2723,21 +2690,8 @@ gdm_slave_session_start (void)
     }
 
     if (ve_string_empty (language)) {
-	    const char *lang = g_getenv ("LANG");
-
 	    g_free (language);
-
-	    if ( ! ve_string_empty (lang))
-		    language = g_strdup (lang);
-	    else
-		    language = g_strdup (GdmDefaultLocale);
-
-	    if (ve_string_empty (language)) {
-		    g_free (language);
-		    language = g_strdup ("C");
-	    }
-
-	    def_language = TRUE;
+	    language = NULL;
     }
 
     /* save this session as the users session */
@@ -2813,13 +2767,9 @@ gdm_slave_session_start (void)
 			     "is not possible to log in.  Please contact\n"
 			     "your system administrator"));
 
-	    gdm_slave_session_stop (0);
-	    gdm_slave_session_cleanup ();
+	    gdm_slave_session_stop (FALSE /* run_post_session */);
 
-	    gdm_server_stop (d);
-	    gdm_verify_cleanup (d);
-
-	    _exit (DISPLAY_REMANAGE);
+	    gdm_slave_quick_exit (DISPLAY_REMANAGE);
     }
 
     if (strcmp (session, GDM_SESSION_FAILSAFE_GNOME) == 0 ||
@@ -2854,7 +2804,6 @@ gdm_slave_session_start (void)
 			   session,
 			   save_session,
 			   language,
-			   def_language,
 			   gnome_session,
 			   usrcfgok,
 			   savesess,
@@ -2919,13 +2868,13 @@ gdm_slave_session_start (void)
 
     gdm_debug ("gdm_slave_session_start: Session ended OK");
 
-    gdm_slave_session_stop (pid);
-    gdm_slave_session_cleanup ();
+    gdm_slave_session_stop (pid != 0);
 }
 
 
+/* Stop any in progress sessions */
 static void
-gdm_slave_session_stop (pid_t sesspid)
+gdm_slave_session_stop (gboolean run_post_session)
 {
     struct passwd *pwent;
     char *x_servers_file;
@@ -2946,19 +2895,22 @@ gdm_slave_session_stop (pid_t sesspid)
      * FIXME: Maybe we should waitpid here, note make sure this will
      * not create a hang! */
     gdm_sigchld_block_push ();
-    if (d->sesspid > 0)
+    if (d->sesspid > 1)
 	    kill (- (d->sesspid), SIGTERM);
     gdm_sigchld_block_pop ();
 
     gdm_verify_cleanup (d);
     
-    pwent = getpwnam (local_login);	/* PAM overwrites our pwent */
+    if (local_login == NULL)
+	    pwent = NULL;
+    else
+	    pwent = getpwnam (local_login);	/* PAM overwrites our pwent */
 
     x_servers_file = g_strconcat (GdmServAuthDir,
 				  "/", d->name, ".Xservers", NULL);
 
     /* if there was a session that ran, run the PostSession script */
-    if (sesspid != 0) {
+    if (run_post_session) {
 	    /* setup some env for PostSession script */
 	    gnome_setenv ("DISPLAY", d->name, TRUE);
 	    gnome_setenv ("XAUTHORITY", d->authfile, TRUE);
@@ -2968,7 +2920,7 @@ gdm_slave_session_stop (pid_t sesspid)
 		    gnome_setenv ("REMOTE_HOST", d->hostname, TRUE);
 
 	    /* Execute post session script */
-	    gdm_debug ("gdm_slave_session_cleanup: Running post session script");
+	    gdm_debug ("gdm_slave_session_stop: Running post session script");
 	    gdm_slave_exec_script (d, GdmPostSession, local_login, pwent,
 				   FALSE /* pass_stdout */,
 				   TRUE /* set_parent */);
@@ -2981,36 +2933,24 @@ gdm_slave_session_stop (pid_t sesspid)
     unlink (x_servers_file);
     g_free (x_servers_file);
 
-    if (pwent == NULL) {
-	    return;
+    g_free (local_login);
+
+    if (pwent != NULL) {
+	    /* Remove display from ~user/.Xauthority */
+	    setegid (pwent->pw_gid);
+	    seteuid (pwent->pw_uid);
+
+	    gdm_auth_user_remove (d, pwent->pw_uid);
+
+	    seteuid (0);
+	    setegid (0);
     }
 
-    g_free (local_login);
-    
-    /* Remove display from ~user/.Xauthority */
-    setegid (pwent->pw_gid);
-    seteuid (pwent->pw_uid);
-
-    gdm_auth_user_remove (d, pwent->pw_uid);
-
-    seteuid (0);
-    setegid (0);
-}
-
-static void
-gdm_slave_session_cleanup (void)
-{
-    gdm_debug ("gdm_slave_session_cleanup: on %s", d->name);
-
-    /* kill login if it still exists */
-    g_free (login);
-    login = NULL;
-    
     /* things are going to be killed, so ignore errors */
     XSetErrorHandler (ignore_xerror_handler);
 
     /* Cleanup */
-    gdm_debug ("gdm_slave_session_cleanup: Severing connection");
+    gdm_debug ("gdm_slave_session_stop: Severing connection");
     if (d->dsp != NULL) {
 	    XCloseDisplay (d->dsp);
 	    d->dsp = NULL;
@@ -3020,53 +2960,16 @@ gdm_slave_session_cleanup (void)
 static void
 gdm_slave_term_handler (int sig)
 {
-	sigset_t tmask;
-
 	gdm_in_signal++;
+	gdm_wait_for_ack = FALSE;
 
 	gdm_debug ("gdm_slave_term_handler: %s got TERM/INT signal", d->name);
 
-	/* just for paranoia's sake */
-	seteuid (0);
-	setegid (0);
+	gdm_slave_session_stop (d->logged_in && login != NULL);
 
-	sigemptyset (&tmask);
-	sigaddset (&tmask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &tmask, NULL);  
+	gdm_debug ("gdm_slave_term_handler: Final cleanup");
 
-	if (extra_process > 1) {
-		/* we sigterm extra processes, and we
-		 * don't wait */
-		kill (extra_process, SIGTERM);
-		extra_process = -1;
-	}
-
-	if (d->greetpid != 0) {
-		pid_t pid = d->greetpid;
-		d->greetpid = 0;
-		greet = FALSE;
-		gdm_debug ("gdm_slave_term_handler: Whacking greeter");
-		if (kill (pid, sig) == 0)
-			ve_waitpid_no_signal (pid, 0, 0); 
-	} else if (login != NULL) {
-		gdm_slave_session_stop (d->sesspid);
-		gdm_slave_session_cleanup ();
-	}
-
-	if (d->chooserpid != 0) {
-		pid_t pid = d->chooserpid;
-		d->chooserpid = 0;
-		gdm_debug ("gdm_slave_term_handler: Whacking chooser");
-		if (kill (pid, sig) == 0)
-			ve_waitpid_no_signal (pid, 0, 0); 
-	}
-
-	gdm_debug ("gdm_slave_term_handler: Whacking server");
-
-	gdm_server_stop (d);
-	gdm_verify_cleanup (d);
-
-	_exit (DISPLAY_ABORT);
+	gdm_slave_quick_exit (DISPLAY_ABORT);
 }
 
 /* called on alarms to ping */
@@ -3088,11 +2991,7 @@ gdm_slave_alrm_handler (int sig)
 
 	if (in_ping) {
 		/* darn, the last ping didn't succeed, wipe this display */
-
-		if (login != NULL) {
-			gdm_slave_session_stop (d->sesspid);
-			gdm_slave_session_cleanup ();
-		}
+		gdm_slave_session_stop (d->logged_in && login != NULL);
 
 		gdm_slave_exit (DISPLAY_REMANAGE, 
 				_("Ping to %s failed, whacking display!"),
@@ -3155,16 +3054,9 @@ gdm_slave_child_handler (int sig)
 
 		whack_greeter_fds ();
 
-		/* just for paranoia's sake */
-		seteuid (0);
-		setegid (0);
-
 		/* if greet is TRUE, then the greeter died outside of our
-		 * control really, so clean up and die, something is wrong */
-		gdm_server_stop (d);
-		gdm_verify_cleanup (d);
-
-		/* The greeter is only allowed to pass back these
+		 * control really, so clean up and die, something is wrong
+		 * The greeter is only allowed to pass back these
 		 * exit codes, else we'll just remanage */
 		if (WIFEXITED (status) &&
 		    (WEXITSTATUS (status) == DISPLAY_ABORT ||
@@ -3172,9 +3064,9 @@ gdm_slave_child_handler (int sig)
 		     WEXITSTATUS (status) == DISPLAY_HALT ||
 		     WEXITSTATUS (status) == DISPLAY_SUSPEND ||
 		     WEXITSTATUS (status) == DISPLAY_RESTARTGDM)) {
-			_exit (WEXITSTATUS (status));
+			gdm_slave_quick_exit (WEXITSTATUS (status));
 		} else {
-			_exit (DISPLAY_REMANAGE);
+			gdm_slave_quick_exit (DISPLAY_REMANAGE);
 		}
 	} else if (pid != 0 && pid == d->sesspid) {
 		d->sesspid = 0;
@@ -3184,12 +3076,31 @@ gdm_slave_child_handler (int sig)
 		d->servstat = SERVER_DEAD;
 		d->servpid = 0;
 		gdm_server_wipe_cookies (d);
+		gdm_slave_whack_temp_auth_file ();
 
+		/* if not handled there is no need for further formalities,
+		 * we just have to die */
 		if ( ! d->handled)
-			_exit (DISPLAY_REMANAGE);
+			gdm_slave_quick_exit (DISPLAY_REMANAGE);
+
+		gdm_slave_send_num (GDM_SOP_XPID, 0);
+
+		/* whack the session good */
+		if (d->sesspid > 1) {
+			gdm_slave_send_num (GDM_SOP_SESSPID, 0);
+			kill (- (d->sesspid), SIGTERM);
+		}
+		if (d->greetpid > 1) {
+			gdm_slave_send_num (GDM_SOP_GREETPID, 0);
+			kill (d->greetpid, SIGTERM);
+		}
+		if (d->chooserpid > 1) {
+			gdm_slave_send_num (GDM_SOP_CHOOSERPID, 0);
+			kill (d->chooserpid, SIGTERM);
+		}
 	} else if (pid == extra_process) {
 		/* an extra process died, yay! */
-		extra_process = -1;
+		extra_process = 0;
 	    	extra_status = status;
 	}
     }
@@ -3250,9 +3161,7 @@ gdm_slave_usr2_handler (int sig)
 				if (d->type != TYPE_FLEXI_XNEST &&
 				    d->type != TYPE_FLEXI) {
 					if ( ! d->logged_in) {
-						gdm_server_stop (d);
-						gdm_verify_cleanup (d);
-						_exit (DISPLAY_REMANAGE);
+						gdm_slave_quick_exit (DISPLAY_REMANAGE);
 					} else {
 						remanage_asap = TRUE;
 					}
@@ -3278,49 +3187,25 @@ gdm_slave_xerror_handler (Display *disp, XErrorEvent *evt)
 static gint
 gdm_slave_xioerror_handler (Display *disp)
 {
-	sigset_t tmask;
+	/* Display is all gone */
+	d->dsp = NULL;
 
 	gdm_debug ("gdm_slave_xioerror_handler: I/O error for display %s", d->name);
 
-	/* just for paranoia's sake */
-	seteuid (0);
-	setegid (0);
+	gdm_slave_session_stop (d->logged_in && login != NULL);
 
-	sigemptyset (&tmask);
-	sigaddset (&tmask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &tmask, NULL);  
-
-	if (d->greetpid != 0) {
-		pid_t pid = d->greetpid;
-		d->greetpid = 0;
-		greet = FALSE;
-		if (kill (pid, SIGINT) == 0)
-			ve_waitpid_no_signal (pid, 0, 0); 
-	} else if (login != NULL) {
-		gdm_slave_session_stop (d->sesspid);
-		gdm_slave_session_cleanup ();
-	}
-
-	if (d->chooserpid != 0) {
-		pid_t pid = d->chooserpid;
-		d->chooserpid = 0;
-		if (kill (pid, SIGINT) == 0)
-			ve_waitpid_no_signal (pid, 0, 0); 
-	}
-    
 	gdm_error (_("gdm_slave_xioerror_handler: Fatal X error - Restarting %s"), d->name);
 
-	gdm_server_stop (d);
-	gdm_verify_cleanup (d);
-
 	if ((d->type == TYPE_LOCAL ||
 	     d->type == TYPE_FLEXI) &&
 	    (do_xfailed_on_xio_error ||
 	     d->starttime + 5 >= time (NULL))) {
-		_exit (DISPLAY_XFAILED);
+		gdm_slave_quick_exit (DISPLAY_XFAILED);
 	} else {
-		_exit (DISPLAY_REMANAGE);
+		gdm_slave_quick_exit (DISPLAY_REMANAGE);
 	}
+
+	return 0;
 }
 
 static void
@@ -3422,6 +3307,53 @@ gdm_slave_greeter_ctl_no_ret (char cmd, const char *str)
 	g_free (gdm_slave_greeter_ctl (cmd, str));
 }
 
+static void 
+gdm_slave_quick_exit (gint status)
+{
+    /* just for paranoia's sake */
+    seteuid (0);
+    setegid (0);
+
+    if (d != NULL) {
+	    /* Well now we're just going to kill
+	     * everything including the X server,
+	     * so no need doing XCloseDisplay which
+	     * may just get us an XIOError */
+	    d->dsp = NULL;
+
+	    /* No need to send the PIDS to the daemon
+	     * since we'll just exit cleanly */
+
+	    /* Push and never pop */
+	    gdm_sigchld_block_push ();
+
+	    /* Kill children where applicable */
+	    if (d->greetpid > 1)
+		    kill (d->greetpid, SIGTERM);
+	    d->greetpid = 0;
+
+	    if (d->chooserpid > 1)
+		    kill (d->chooserpid, SIGTERM);
+	    d->chooserpid = 0;
+
+	    if (d->sesspid > 1)
+		    kill (-(d->sesspid), SIGTERM);
+	    d->sesspid = 0;
+
+	    gdm_server_stop (d);
+	    gdm_verify_cleanup (d);
+
+	    if (d->servpid > 1)
+		    kill (d->servpid, SIGTERM);
+	    d->servpid = 0;
+
+	    if (extra_process > 1)
+		    kill (extra_process, SIGTERM);
+	    extra_process = 0;
+    }
+
+    _exit (status);
+}
 
 static void 
 gdm_slave_exit (gint status, const gchar *format, ...)
@@ -3433,43 +3365,11 @@ gdm_slave_exit (gint status, const gchar *format, ...)
     s = g_strdup_vprintf (format, args);
     va_end (args);
     
-    syslog (LOG_ERR, "%s", s);
+    gdm_error ("%s", s);
     
     g_free (s);
 
-    /* just for paranoia's sake */
-    seteuid (0);
-    setegid (0);
-
-    if (d != NULL) {
-	    sigset_t tmask;
-
-	    sigemptyset (&tmask);
-	    sigaddset (&tmask, SIGCHLD);
-	    sigprocmask (SIG_BLOCK, &tmask, NULL);  
-
-	    /* Kill children where applicable */
-	    if (d->greetpid != 0)
-		    kill (d->greetpid, SIGTERM);
-	    d->greetpid = 0;
-
-	    if (d->chooserpid != 0)
-		    kill (d->chooserpid, SIGTERM);
-	    d->chooserpid = 0;
-
-	    if (d->sesspid != 0)
-		    kill (-(d->sesspid), SIGTERM);
-	    d->sesspid = 0;
-
-	    gdm_server_stop (d);
-	    gdm_verify_cleanup (d);
-
-	    if (d->servpid != 0)
-		    kill (d->servpid, SIGTERM);
-	    d->servpid = 0;
-    }
-
-    _exit (status);
+    gdm_slave_quick_exit (status);
 }
 
 static void 
@@ -3835,11 +3735,11 @@ gdm_slave_handle_notify (const char *msg)
 		GdmAllowRemoteAutoLogin = val;
 	} else if (sscanf (msg, GDM_NOTIFY_SYSMENU " %d", &val) == 1) {
 		GdmSystemMenu = val;
-		if (d->greetpid > 0)
+		if (d->greetpid > 1)
 			kill (d->greetpid, SIGHUP);
 	} else if (sscanf (msg, GDM_NOTIFY_CONFIG_AVAILABLE " %d", &val) == 1) {
 		GdmConfigAvailable = val;
-		if (d->greetpid > 0)
+		if (d->greetpid > 1)
 			kill (d->greetpid, SIGHUP);
 	} else if (sscanf (msg, GDM_NOTIFY_RETRYDELAY " %d", &val) == 1) {
 		GdmRetryDelay = val;
@@ -3855,9 +3755,7 @@ gdm_slave_handle_notify (const char *msg)
 				/* FIXME: can't handle flexi servers like this
 				 * without going all cranky */
 				if ( ! d->logged_in) {
-					gdm_server_stop (d);
-					gdm_verify_cleanup (d);
-					_exit (DISPLAY_REMANAGE);
+					gdm_slave_quick_exit (DISPLAY_REMANAGE);
 				} else {
 					remanage_asap = TRUE;
 				}
@@ -3875,9 +3773,7 @@ gdm_slave_handle_notify (const char *msg)
 				/* FIXME: can't handle flexi servers like this
 				 * without going all cranky */
 				if ( ! d->logged_in) {
-					gdm_server_stop (d);
-					gdm_verify_cleanup (d);
-					_exit (DISPLAY_REMANAGE);
+					gdm_slave_quick_exit (DISPLAY_REMANAGE);
 				} else {
 					remanage_asap = TRUE;
 				}
@@ -3889,16 +3785,14 @@ gdm_slave_handle_notify (const char *msg)
 		/* FIXME: can't handle flexi servers without going all cranky */
 		if (d->type == TYPE_LOCAL || d->type == TYPE_XDMCP) {
 			if ( ! d->logged_in) {
-				gdm_server_stop (d);
-				gdm_verify_cleanup (d);
-				_exit (DISPLAY_REMANAGE);
+				gdm_slave_quick_exit (DISPLAY_REMANAGE);
 			} else {
 				remanage_asap = TRUE;
 			}
 		}
 	} else if (sscanf (msg, GDM_NOTIFY_TIMED_LOGIN_DELAY " %d", &val) == 1) {
 		GdmTimedLoginDelay = val;
-		if (d->greetpid > 0)
+		if (d->greetpid > 1)
 			kill (d->greetpid, SIGHUP);
 	}
 }
