@@ -48,7 +48,6 @@ static pam_handle_t *pamh = NULL;
 
 static GdmDisplay *cur_gdm_disp = NULL;
 
-
 /* Internal PAM conversation function. Interfaces between the PAM
  * authentication system and the actual greeter program */
 
@@ -131,13 +130,16 @@ static struct pam_conv pamc = {
     NULL
 };
 
+/* Extra message to give on queries */
+static char *extra_standalone_message = NULL;
+
 static gint 
 gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 				struct pam_response **resp,
 				void *appdata_ptr)
 {
 	int replies = 0;
-	char *s;
+	char *s, *text;
 	struct pam_response *reply = NULL;
 
 	reply = malloc (sizeof (struct pam_response) * num_msg);
@@ -151,10 +153,17 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 		switch (msg[replies]->msg_style) {
 
 		case PAM_PROMPT_ECHO_ON:
+			if (extra_standalone_message != NULL)
+				text = g_strdup_printf
+					("%s\n%s", extra_standalone_message,
+					 _(msg[replies]->msg));
+			else
+				text = g_strdup (_(msg[replies]->msg));
+
 			/* PAM requested textual input with echo on */
-			s = gdm_failsafe_question (cur_gdm_disp,
-						   _(msg[replies]->msg),
+			s = gdm_failsafe_question (cur_gdm_disp, text,
 						   TRUE /* echo */);
+			g_free (text);
 
 			reply[replies].resp_retcode = PAM_SUCCESS;
 			reply[replies].resp = strdup (ve_sure_string (s));
@@ -162,11 +171,19 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 			break;
 
 		case PAM_PROMPT_ECHO_OFF:
+			if (extra_standalone_message != NULL)
+				text = g_strdup_printf
+					("%s\n%s", extra_standalone_message,
+					 _(msg[replies]->msg));
+			else
+				text = g_strdup (_(msg[replies]->msg));
+
 			/* PAM requested textual input with echo off */
-			s = gdm_failsafe_question (cur_gdm_disp,
-						   _(msg[replies]->msg),
+			s = gdm_failsafe_question (cur_gdm_disp, text,
 						   FALSE /* echo */);
 
+			g_free (text);
+
 			reply[replies].resp_retcode = PAM_SUCCESS;
 			reply[replies].resp = strdup (ve_sure_string (s));
 			g_free (s);
@@ -282,12 +299,13 @@ gdm_verify_user (GdmDisplay *d,
 		 gboolean local) 
 {
     gint pamerr = 0;
-    gchar *login;
+    char *login;
+    const char *service = "gdm";
     struct passwd *pwent;
     gboolean error_msg_given = FALSE;
     gboolean credentials_set = FALSE;
     gboolean started_timer = FALSE;
-    gchar *auth_errmsg;
+    char *auth_errmsg;
 
     /* start the timer for timed logins */
     if ( ! ve_string_empty (GdmTimedLogin) &&
@@ -312,10 +330,17 @@ gdm_verify_user (GdmDisplay *d,
 	    login = g_strdup (username);
     }
 
+    if (local &&
+	gdm_is_a_no_password_user (login)) {
+	    service = "gdm-autologin";
+    } else {
+	    service = "gdm";
+    }
+
     cur_gdm_disp = d;
 
     /* Initialize a PAM session for the user */
-    if ( ! create_pamh (d, "gdm", login, &pamc, display, &pamerr)) {
+    if ( ! create_pamh (d, service, login, &pamc, display, &pamerr)) {
 	    if (started_timer)
 		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 	    goto pamerr;
@@ -490,6 +515,11 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 
     cur_gdm_disp = d;
 
+    g_free (extra_standalone_message);
+    extra_standalone_message = g_strdup_printf ("%s (%s)",
+						_("Automatic login"),
+						login);
+
     /* Initialize a PAM session for the user */
     if ( ! create_pamh (d, "gdm-autologin", login, &standalone_pamc,
 			display, &pamerr)) {
@@ -582,6 +612,9 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
     openlog ("gdm", LOG_PID, LOG_DAEMON);
 
     cur_gdm_disp = NULL;
+
+    g_free (extra_standalone_message);
+    extra_standalone_message = NULL;
     
     return TRUE;
     
@@ -597,6 +630,9 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 
     cur_gdm_disp = NULL;
 
+    g_free (extra_standalone_message);
+    extra_standalone_message = NULL;
+
     return FALSE;
 }
 
