@@ -1,3 +1,5 @@
+#include "config.h"
+
 #include <gtk/gtk.h>
 #include <libxml/parser.h>
 #include <string.h>
@@ -8,6 +10,8 @@
 #include <locale.h>
 #include <libgnome/gnome-i18n.h>
 
+#include <vicious.h>
+
 #include "greeter_parser.h"
 #include "greeter_events.h"
 
@@ -864,7 +868,7 @@ parse_translated_text (xmlNodePtr node,
   xmlChar *prop;
   gint score;
   
-  prop = xmlGetProp (node, "xml:lang");
+  prop = xmlNodeGetLang (node);
   if (prop)
     {
       score = is_current_locale (prop);
@@ -957,6 +961,14 @@ parse_label (xmlNodePtr        node,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
 		   "A label must specify the text attribute");
     }
+  /* FIXME: evil hack to use internally translated strings */
+  if (translation_score == 999 &&
+      ! ve_string_empty (translated_text))
+    {
+      char *foo = g_strdup (_(translated_text));
+      g_free (translated_text);
+      translated_text = foo;
+    }
 
   for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
     {
